#pragma rtGlobals=1		// Use modern global access method.
#pragma IgorVersion=6.01
#pragma version = 1.0
#pragma IndependentModule=ACL_CreateLayouts
#pragma ModuleName=ACL_CreateLayouts

// ********************
//  LICENSE
// ********************
//	Copyright (c) 2007 by Adam Light
//	
//	Permission is hereby granted, free of charge, to any person
//	obtaining a copy of this software and associated documentation
//	files (the "Software"), to deal in the Software without
//	restriction, including without limitation the rights to use,
//	copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the
//	Software is furnished to do so, subject to the following
//	conditions:
//	
//	The above copyright notice and this permission notice shall be
//	included in all copies or substantial portions of the Software.
//	
//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//	EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
//	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
//	NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
//	HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
//	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
//	FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
//	OTHER DEALINGS IN THE SOFTWARE.
//
// *************************
//  VERSION HISTORY
// *************************
//	Date			Version #				Changes
//	Jun 21, 2007		Version 1.0.0.0			Initial release


// ********************
//  CONSTANTS
// ********************
// module name--use the same name as in the '#pragma ModuleName= ' declaration
static StrConstant kMODULENAME = "ACL_CreateLayouts"

// package data folder
static StrConstant kPACKAGEDF = "root:Packages:ACL_CreateLayouts"

// preferences saving
static StrConstant kPackageName = "ACL_CreateLayouts"
static StrConstant kPreferencesFileName = "ACL_CreateLayoutsPrefs.bin"
static Constant kPreferencesRecordID = 0
static Constant K_CURRENTPREFSVERSION = 100

// help text string
static StrConstant kHELPTEXTSTRING = "Create Layout GUI"

// Main panel
static StrConstant kMAINPANELNAME = "CreateLayoutsPanel"
static StrConstant kMAINPANELTITLE = "Create Layout"
static Constant kPANEL_DEFAULT_HEIGHT = 436			// in pixels
static Constant kPANEL_DEFAULT_WIDTH = 740			// in pixels
static Constant kPANEL_MIN_HEIGHT = 300				// in pixels
static Constant kPANEL_MIN_LEFT_COL_WIDTH = 340		// in pixels

// Repositioning/resizing constants.
static Constant kNO_RESIZE = 0
static Constant kWIDTH_RESIZE = 1
static Constant kHEIGHT_RESIZE = 2
static Constant kBOTH_RESIZE = 3

// Row heights of icons in list boxes, in pixels.
static Constant kMIN_LB_ICON_HEIGHT = 100
static Constant kMAX_LB_ICON_HEIGHT = 350


// Igor 6.10B03 included a change so that tables are supported
// when using ListBox special kind=1.
// Types of windows to include in panel.  1=graphs only, 2=tables only, 3=graphs and tables
#if NumberByKey("IGORVERS", IgorInfo(0)) >= 6.1
	static Constant kWINDOWTYPES = 3	// Graphs and tables.
#else
	static Constant kWINDOWTYPES = 1	// Graphs only.
#endif

// ********************
//  STRUCTURES
// ********************
Structure ACL_CreateLayoutsStr
	uint32 version
	double panelCoords[4]		// left, top, right, bottom
	uint32 reserved[100]		// reserved for future use
EndStructure

Structure createLayoutGlobals
	WAVE/T windowListText
	WAVE windowListSelect
	WAVE/T positionListText
	WAVE positionListSelect
	WAVE windowListColors
	WAVE/T stringTextWave
	WAVE/T selectedWindows
	WAVE/T iconListText
	NVAR orientation
	NVAR stackTile
	NVAR groutThickness
	NVAR layoutRows
	NVAR layoutCols
	NVAR usePrefs
	NVAR frameStyleMode
EndStructure

// ********************
//  GLOBAL VARIABLES
// ********************
Function GetCreateLayoutsGlobals(cl, pName)		// also creates globals if they do not yet exist
	// this function modified from GraphBrowser.ipf
	STRUCT createLayoutGlobals &cl
	String pName		// panel name
	
	String currentDF = GetDataFolder(1)
	NewDataFolder/O/S root:Packages
	NewDataFolder/O/S $(kPACKAGEDF)
	
	if (cmpstr(pName, kMAINPANELNAME) != 0)	// working on an additional panel beyond the base?
		NewDataFolder/O/S $(pName)
	endif
	
	WAVE/Z/T cl.windowListText
	if (!WaveExists(cl.windowListText))
		// main list box text and select waves
		Variable numCols = 5
		Make/O/T/N=(1, numCols) windowListText
		Make/O/N=(1,numCols, 3) windowListSelect
		SetDimLabel 1, 0,Selected, windowListText, windowListSelect
		SetDimLabel 1, 1,Name, windowListText, windowListSelect
		SetDimLabel 1, 2,Type, windowListText, windowListSelect
		SetDimLabel 1, 3,TypeCode, windowListText, windowListSelect
		SetDimLabel 2, 1,foreColors, windowListSelect
		SetDimLabel 2, 2,backColors, windowListSelect
		Redimension/N=(0, -1, -1) windowListText, windowListSelect
		
		// position list box text and select waves
		Make/O/T/N=(1,0) positionListText
		Make/O/N=(1,1,3) positionListSelect
		SetDimLabel 2, 1,foreColors, positionListSelect
		SetDimLabel 2, 2,backColors, positionListSelect
		Redimension/N=(0, 0, -1) positionListText, positionListSelect
		// make color wave
		Make/O/W/U windowListColors = {{65535, 65535, 65535}, {51143, 64250, 64764}}
		MatrixTranspose windowListColors
		// selected windows wave
		Make/O/T/N=(0) selectedWindows
		// one cell text box that contains graph icon used during drag and drop
		Make/O/T/N=1 iconListText
		
		// make text wave to store string variables
		Make/O/T/N=(10) stringTextWave
		SetDimLabel 0, 0, LayoutName, stringTextWave
		SetDimLabel 0, 1, LayoutTitle, stringTextWave
		
		Variable/G orientation = 1		// 0=portrait, 1=landscape
		Variable/G stackTile = 0		// 0=neither, 1=stack, 2=tile
		Variable/G groutThickness = 5	// in pixels
		Variable/G layoutRows = 2
		Variable/G layoutCols = 2
		Variable/G usePrefs = 1		// should layout be created using user's preferences or not
		
		Variable/G frameStyleMode = 2	// single is default

	endif
	
	WAVE/T cl.windowListText
	WAVE cl.windowListSelect
	WAVE/T cl.positionListText
	WAVE cl.positionListSelect
	WAVE cl.windowListColors
	WAVE/T cl.stringTextWave
	WAVE/T cl.selectedWindows
	WAVE/T cl.iconListText
	NVAR cl.orientation
	NVAR cl.stackTile
	NVAR cl.groutThickness
	NVAR cl.layoutRows
	NVAR cl.layoutCols
	NVAR cl.usePrefs
	NVAR cl.frameStyleMode
	SetDataFolder currentDF
End


// ********************
//  MENUS
// ********************
Menu "Misc", hideable
	"Create Layout",/Q,Initialize_pnlMain()
End

// ********************
//  PANEL
// ********************
// To update or change this, edit the recreation macro like so:
// Edit the first lines to match
// serch for "root:Packages:ACL_CreateLayouts:" and replace with "g."
Function NewCreateLayoutPanel(pName, x0, y0)
	// modified from NewGraphBrowserPanel() in GraphBrowser.ipf by WaveMetrics
	String pName
	Variable x0, y0		// if non-zero, this is taken to be the origin
	
	if( x0==0 && y0==0 )
		x0= 450
		y0= 60
	endif
	
	STRUCT ACL_CreateLayoutsStr prefs
	LoadDesktopControlPrefs(prefs)
			
	NewPanel /W=(x0, y0,x0 + kPANEL_DEFAULT_WIDTH,y0 + kPANEL_DEFAULT_HEIGHT)/K=1/N=$(pName) as kMAINPANELTITLE
	ModifyPanel noEdit = 1	// set K=0 if this is commented out
	
	STRUCT createLayoutGlobals cl
	GetCreateLayoutsGlobals(cl, pName)

	ListBox listWindows,pos={10,32},size={356,286},proc=ACL_CreateLayouts#ListBox_Main_Action
	ListBox listWindows,listWave=root:Packages:ACL_CreateLayouts:windowListText
	ListBox listWindows,selWave=root:Packages:ACL_CreateLayouts:windowListSelect
	ListBox listWindows,colorWave=root:Packages:ACL_CreateLayouts:windowListColors
	ListBox listWindows,mode= 2,selRow= 0,special= {1,120,0},widths={40,173,0,0,0}
	ListBox listWindows,userColumnResize= 1
	Button buttonCreateLayout,pos={390,406},size={50,20},proc=ACL_CreateLayouts#ButtonCreateLayout,title="Do It"
	Button buttonCancel,pos={460,406},size={50,20},proc=ACL_CreateLayouts#ButtonCancelLayout,title="Cancel"
	SetVariable setvar0,pos={11,340},size={212,16},proc=ACL_CreateLayouts#SetVarLayoutName,title="Name"
	SetVariable setvar0,value= root:Packages:ACL_CreateLayouts:stringTextWave[%LayoutName],bodyWidth= 180
	SetVariable setvarTitle,pos={10,360},size={213,16},title="Title   ",proc=ACL_CreateLayouts#SetVarTitle
	SetVariable setvarTitle,value= root:Packages:ACL_CreateLayouts:stringTextWave[%LayoutTitle],bodyWidth= 180
	GroupBox groupOrientation,pos={237,333},size={112,67},title="Page Orientation"
	CheckBox checkOrientation_port,pos={249,358},size={51,14},proc=ACL_CreateLayouts#CheckOrientation,title="Portrait"
	CheckBox checkOrientation_port,value= 0,mode=1
	CheckBox checkOrientation_land,pos={249,377},size={71,14},proc=ACL_CreateLayouts#CheckOrientation,title="Landscape"
	CheckBox checkOrientation_land,value= 1,mode=1
	CheckBox checkTile,pos={10,384},size={35,14},proc=ACL_CreateLayouts#CheckStackTile,title="Tile"
	CheckBox checkTile,value= 0
	CheckBox checkStack,pos={59,384},size={46,14},proc=ACL_CreateLayouts#CheckStackTile,title="Stack"
	CheckBox checkStack,value= 0
	SetVariable setvarGroutThickness,pos={115,384},size={90,16},title="Grout"
	SetVariable setvarGroutThickness,limits={0,1000,5},value= root:Packages:ACL_CreateLayouts:groutThickness,bodyWidth= 60
	SetVariable setvarLayoutRows,pos={390,370},size={71,16},proc=ACL_CreateLayouts#SetVarLayoutRowColNum,title="Rows"
	SetVariable setvarLayoutRows,format="%d"
	SetVariable setvarLayoutRows,limits={0,30,1},value= root:Packages:ACL_CreateLayouts:layoutRows,bodyWidth= 40
	SetVariable setvarLayoutCols,pos={481,370},size={84,16},proc=ACL_CreateLayouts#SetVarLayoutRowColNum,title="Columns"
	SetVariable setvarLayoutCols,format="%d"
	SetVariable setvarLayoutCols,limits={0,30,1},value= root:Packages:ACL_CreateLayouts:layoutCols,bodyWidth= 40
	ListBox listPosition,pos={390,35},size={340,325},proc=ACL_CreateLayouts#ListBox_Position_Action
	ListBox listPosition,listWave=root:Packages:ACL_CreateLayouts:positionListText
	ListBox listPosition,selWave=root:Packages:ACL_CreateLayouts:positionListSelect
	ListBox listPosition,colorWave=root:Packages:ACL_CreateLayouts:windowListColors
	ListBox listPosition,mode= 6,special= {1,120,0}
	TitleBox titleWindowSelection,pos={127,7},size={123,16},title="Window Selection"
	TitleBox titleWindowSelection,fSize=14,frame=0,fStyle=1
	TitleBox titlePositioning,pos={518,7},size={77,16},title="Positioning",fSize=14
	TitleBox titlePositioning,frame=0,fStyle=1
	PopupMenu popupFrameStyle,pos={10,406},size={100,21},proc=ACL_CreateLayouts#PopMenuFrameStyle,title="Frame"
	PopupMenu popupFrameStyle,mode=2,popvalue="Single",value= #"\"None;Single;Double;Triple;Shadow\""
	CheckBox checkUsePrefs,pos={145,409},size={132,14},title="Use Layout Preferences"
	CheckBox checkUsePrefs,variable= root:Packages:ACL_CreateLayouts:usePrefs
	Button buttonSelectAll,pos={10,9},size={25,20},title="All",proc=ACL_CreateLayouts#Button_SelectAll
	Button buttonSelectNone,pos={40,9},size={35,20},title="None",proc=ACL_CreateLayouts#Button_SelectNone
	
	// this is the icon list box that I'm testing out--it doesn't work well yet because the scroll bars on the right are always
	// present
//	ListBox listIconDrag,pos={155,421},size={170,130},frame=0
//	ListBox listIconDrag,listWave=cl.iconListText
//	ListBox listIconDrag,special= {1,120,0}
	
End

// ********************
//  INITIALIZATION
// ********************
Function Initialize_pnlMain()
	String panelName = kMAINPANELNAME
	DoWindow panelName
	if (!V_flag)
		String curDataFolder = GetDataFolder(1)
		STRUCT createLayoutGlobals cl
		GetCreateLayoutsGlobals(cl, panelName)
		
		// set a default name for the new layout
		String defaultLayoutName = UniqueName("Layout", 8, 0)
		if (cmpstr(cl.stringTextWave[%LayoutName], "") == 0)
			cl.stringTextWave[%LayoutName] = defaultLayoutName
		endif
		
		BuildWindowListWaves("*", kWINDOWTYPES, cl)				
		NewCreateLayoutPanel(panelName, 0, 0)
		SetWindow $(panelName) hook($(kPackageName))=Hook_MainPanel
		STRUCT ACL_CreateLayoutsStr prefs
		LoadDesktopControlPrefs(prefs)
		MoveWindow/W=$(panelName) prefs.panelCoords[0], prefs.panelCoords[1], prefs.panelCoords[2], prefs.panelCoords[3] 
		ParamChange_RowCol(cl)			// update layout # rows and column stuff
		ParamChange_SelectedWindows(cl)	// update order of windows in selectedWindows wave
		SetCheckboxes(cl, panelName)		// set the checked state of the checkboxes depending on the value of NVARs
		SetDataFolder curDataFolder	
	endif
End

// ********************
//  ACTION FUNCTIONS
// ********************
Function ButtonHelp(ba) : ButtonControl
	STRUCT WMButtonAction &ba

	switch( ba.eventCode )
		case 2: // mouse up
			DisplayHelpTopic/Z kHELPTEXTSTRING
			if (V_Flag != 0)	// help topic not found
				// get the name of this procedure file
				String alertString
				sprintf alertString, "The %s.ihf help file could not be found.  It should be in the same directory as the %s.ipf procedure file.", kPackageName, kPackageName
				DoAlert 0, alertString
			endif
			break
	endswitch

	return 0
End


Function Button_SelectNone(ba) : ButtonControl
	STRUCT WMButtonAction &ba
	
	STRUCT createLayoutGlobals cl
	GetCreateLayoutsGlobals(cl, ba.win)
	
	Variable n, numWindows
	
	switch( ba.eventCode )
		case 2: // mouse up
			cl.windowListSelect[][%Selected][0] = cl.windowListSelect[p][q][r] & ~0x10		// de-select all rows
			STRUCT WMListboxAction lba
			numWindows = DimSize(cl.windowListSelect, 0)
			lba.col = 0
			lba.win = ba.win
			WAVE/T lba.listWave = cl.windowListText
			WAVE lba.selWave = cl.windowListSelect
			For (n=0; n<numWindows; n+=1)
				lba.row = n
				HandleWindowSelection(lba, cl)
			EndFor
			break
	endswitch

	return 0
End

Function Button_SelectAll(ba) : ButtonControl
	STRUCT WMButtonAction &ba

	STRUCT createLayoutGlobals cl
	GetCreateLayoutsGlobals(cl, ba.win)
	
	Variable numWindows, n
	switch( ba.eventCode )
		case 2: // mouse up
			cl.windowListSelect[][%Selected][0] = cl.windowListSelect[p][q][r] | 0x10		// select all rows
			STRUCT WMListboxAction lba
			numWindows = DimSize(cl.windowListSelect, 0)
			lba.col = 0
			lba.win = ba.win
			WAVE/T lba.listWave = cl.windowListText
			WAVE lba.selWave = cl.windowListSelect
			For (n=0; n<numWindows; n+=1)
				lba.row = n
				HandleWindowSelection(lba, cl)
			EndFor
			break
	endswitch

	return 0
End

Function ButtonCreateLayout(ba) : ButtonControl
	STRUCT WMButtonAction &ba

	STRUCT createLayoutGlobals cl
	GetCreateLayoutsGlobals(cl, ba.win)
	
	switch( ba.eventCode )
		case 2: // mouse up
			CreateLayout(ba.win)
			break
	endswitch

	return 0
End


Function ButtonCancelLayout(ba) : ButtonControl
	STRUCT WMButtonAction &ba
	
	STRUCT createLayoutGlobals cl
	GetCreateLayoutsGlobals(cl, ba.win)
	
	switch( ba.eventCode )
		case 2: // mouse up
			Kill_CreateLayouts(cl, ba.win)		// cleanup and kill panel
			break
	endswitch

	return 0
End

Function SetVarLayoutName(sva) : SetVariableControl
	STRUCT WMSetVariableAction &sva

	switch( sva.eventCode )
		case 1: // mouse up
		case 2: // Enter key
		case 3: // Live update
			String sval = sva.sval
			WAVE/T textWave = sva.svWave
			
			if (cmpstr(textWave[sva.rowIndex][sva.colIndex], "") == 0)
				textWave[sva.rowIndex][sva.colIndex] = "Layout"
			endif
			
			// clean up name
			textWave[sva.rowIndex][sva.colIndex] = CleanupName(textWave[sva.rowIndex][sva.colIndex], 0)
		
			// make sure the name of the layout is not already in use
			String uniqueNameBase = textWave[sva.rowIndex][sva.colIndex]
			if (CheckName(textWave[sva.rowIndex][sva.colIndex], 8) != 0)		// name is not unique
				textWave[sva.rowIndex][sva.colIndex] = UniqueName(uniqueNameBase, 8, 0)				
			endif
			break
	endswitch

	return 0
End

Function SetVarTitle(sva) : SetVariableControl
	STRUCT WMSetVariableAction &sva

	switch( sva.eventCode )
		case 1: // mouse up
		case 2: // Enter key
		case 3: // Live update
			String sval = sva.sval
			WAVE/T textWave = sva.svWave
			
			if (strlen(sval) > 40)		// Layout titles are apparently limited to 40 characters
				textWave[sva.rowIndex][sva.colIndex] = sval[0,39]
				DoAlert 0, "Layout titles must be 40 characters or less.  The specified title has been truncated to fit within this limitation."
			endif
			break
	endswitch

	return 0
End

Function SetVarLayoutRowColNum(sva) : SetVariableControl
	STRUCT WMSetVariableAction &sva

	switch( sva.eventCode )
		case 1: // mouse up
		case 2: // Enter key
		case 3: // Live update
			STRUCT createLayoutGlobals cl
			GetCreateLayoutsGlobals(cl, sva.win)
			ParamChange_RowCol(cl)
			
			// Change the row height of list box displaying icons, if necessary.
			RecalculateListBoxRowHeight(sva.win)
			break
	endswitch

	return 0
End

Function CheckOrientation(cba) : CheckBoxControl
	STRUCT WMCheckboxAction &cba

	switch( cba.eventCode )
		case 2: // mouse up
			STRUCT createLayoutGlobals cl
			GetCreateLayoutsGlobals(cl, cba.win)
			
			StrSwitch (cba.ctrlName)
				Case "checkOrientation_port":
					if (cba.checked)
						cl.orientation = 0
						CheckBox checkOrientation_land win=$(cba.win), value=0
					else
						cl.orientation = 1
						CheckBox checkOrientation_land win=$(cba.win), value=1
					endif
					break
				Case "checkOrientation_land":
					if (cba.checked)
						cl.orientation = 1
						CheckBox checkOrientation_port win=$(cba.win), value=0
					else
						cl.orientation = 0
						CheckBox checkOrientation_port win=$(cba.win), value=1
					endif
					break			
			EndSwitch
			break
	endswitch

	return 0
End

Function CheckStackTile(cba) : CheckBoxControl
	STRUCT WMCheckboxAction &cba
	switch( cba.eventCode )
		case 2: // mouse up
			STRUCT createLayoutGlobals cl
			GetCreateLayoutsGlobals(cl, cba.win)
			Variable oppositeChecked = 0
			StrSwitch (cba.ctrlName)
				Case "checkStack":
					// get state of checkTile control box
					ControlInfo/W=$(cba.win) checkTile
					if (V_Flag == 2)
						oppositeChecked = V_value
					else
						return 0
					endif
					if (cba.checked)
						cl.stackTile = 1
						CheckBox checkTile win=$(cba.win), value=0
					elseif (oppositeChecked)
						cl.stackTile = 2
					else
						cl.stackTile = 0
					endif
					break
				Case "checkTile":
					// get state of checkStack control box
					ControlInfo/W=$(cba.win) checkStack
					if (V_Flag == 2)
						oppositeChecked = V_value
					else
						return 0
					endif
					if (cba.checked)
						cl.stackTile = 2
						CheckBox checkStack win=$(cba.win), value=0
					elseif (oppositeChecked)
						cl.stackTile = 1
					else
						cl.stackTile = 0
					endif					
					break
			EndSwitch
			break
	endswitch
	return 0
End

Function ListBox_Main_Action(lba) : ListBoxControl
	STRUCT WMListboxAction &lba
	
	STRUCT createLayoutGlobals cl
	GetCreateLayoutsGlobals(cl, lba.win)

	Variable row = lba.row
	Variable col = lba.col
	WAVE/T/Z listWave = lba.listWave
	WAVE/Z selWave = lba.selWave

	switch( lba.eventCode )
		case -1: // control being killed
			break
		case 1:	// mouse down
			break
		case 2:	// mouse up (also space bar on a check box cell)
			if (lba.row >= 0)		// ignore any clicks in the title bar area
				if (cmpstr(GetDimLabel(lba.selWave, 1, lba.col), "Selected") == 0)		// if column is the "Selected" column
					HandleWindowSelection(lba, cl)
				endif
			endif
			break
		case 3: // double click
			break
		case 4: // cell selection
		case 5: // cell selection plus shift key
			break
		case 6: // begin edit
			break
		case 7: // finish edit
			break
	endswitch

	return 0
End

Function ListBox_Position_Action(lba) : ListBoxControl
	STRUCT WMListboxAction &lba
	
	STRUCT createLayoutGlobals cl
	GetCreateLayoutsGlobals(cl, lba.win)

	Variable row = lba.row
	Variable col = lba.col
	WAVE/T/Z listWave = lba.listWave
	WAVE/Z selWave = lba.selWave

	switch( lba.eventCode )
		case -1: // control being killed
			break
		case 1:	// mouse down
			ListBox $(lba.ctrlName) win=$(lba.win), userdata(startDragRow)=num2str(lba.row), userdata(startDragCol)=num2str(lba.col)
			cl.iconListText = cl.positionListText[lba.row][lba.col]		// put picture into the icon list box
			break
		case 2:	// mouse up (also space bar on a check box cell)
			Variable startDragRow, startDragCol
			startDragRow = str2num(GetUserData(lba.win, lba.ctrlName, "startDragRow"))
			startDragCol = str2num(GetUserData(lba.win, lba.ctrlName, "startDragCol"))
			if (numtype(startDragRow) == 0 && numtype(startDragCol) == 0)
				if (startDragRow != lba.row || startDragCol != lba.col)
					// rearrange the graph order withing the wave as a consequence of drag and drop
					Variable startPoint1D, endPoint1D
					startPoint1D = (startDragRow * DimSize(lba.listWave, 1)) + startDragCol
					endPoint1D = (lba.row * DimSize(lba.listWave, 1)) + lba.col
					Variable numGraphs = DimSize(cl.selectedWindows, 0)
					if (endPoint1D < numGraphs)		// don't move graph if end point is an empty cell
						if (startPoint1D > endPoint1D)
							startPoint1D += 1
							InsertPoints/M=0 endPoint1D, 1, cl.selectedWindows
							cl.selectedWindows[endPoint1D] = cl.selectedWindows[startPoint1D]
							DeletePoints/M=0 startPoint1D, 1, cl.selectedWindows
						else
							InsertPoints/M=0 endPoint1D + 1, 1, cl.selectedWindows
							cl.selectedWindows[endPoint1D + 1] = cl.selectedWindows[startPoint1D]
							DeletePoints/M=0 startPoint1D, 1, cl.selectedWindows						
						endif
						ParamChange_SelectedWindows(cl)
					endif
				endif
			endif
			break
		case 3: // double click
			break
		case 4: // cell selection
		case 5: // cell selection plus shift key
			break
		case 6: // begin edit
			break
		case 7: // finish edit
			break
	endswitch

	return 0
End

Function PopMenuFrameStyle(pa) : PopupMenuControl
	STRUCT WMPopupAction &pa

	STRUCT createLayoutGlobals cl
	GetCreateLayoutsGlobals(cl, pa.win)
	
	switch( pa.eventCode )
		case 2: // mouse up
			cl.frameStyleMode = pa.popNum
			break
	endswitch

	return 0
End

// ********************
//  AUXILLARY FUNCTIONS
// ********************
static Function BuildWindowListWaves(matchStr, findTypeCode, cl)
	// makes text and select waves with available windows (graphs and/or tables) that can be appended to a layout
	String matchStr
	Variable findTypeCode	
	STRUCT createLayoutGlobals &cl

	
	String windowList = WinList(matchStr, ";", "WIN:" + num2str(findTypeCode))
	Variable n, numWindows, typeCode
	String currentName, typeName
	numWindows = ItemsInList(windowList, ";")
	
	Redimension/N=(numWindows, -1, -1) cl.windowListText, cl.windowListSelect
	For (n=0; n<numWindows; n+=1)
		currentName = StringFromList(n, windowList, ";")
		cl.windowListText[n][%Name] = currentName
		typeCode = WinType(currentName)
		cl.windowListText[n][%TypeCode] = num2str(typeCode)
		Switch (typeCode)
			Case 1:		// graph
				cl.windowListText[n][%Type] = "Graph"
				break
			Case 2:		// table
				cl.windowListText[n][%Type] = "Table"
				break
			default:
		EndSwitch
		
		// set select wave options
		cl.windowListSelect[n][%Selected][0] = cl.windowListSelect[p][q][r] | 0x20		// make it a checkbox
		cl.windowListSelect[n][%Type][0] = cl.windowListSelect[p][q][r] | 0x20		// make it a checkbox
		
		// set background colors to be alternating for odd/even rows
		cl.windowListSelect[n][][%backColors] = mod(p,2)
	EndFor
End

static Function Kill_CreateLayouts(cl, panelName)
	// cleans up when the user cancels or when the panel should otherwise be killed
	STRUCT createLayoutGlobals &cl
	String panelName
	
	DoWindow/K $(panelName)
End

static Function SetCheckboxes(cl, panelName)		// set the checked state of the checkboxes depending on the value of NVARs
	STRUCT createLayoutGlobals &cl
	String panelName
	
	// tile/stack checkboxes	
	Variable stackWin, tileWin
	if (cl.stackTile == 1)		// stack
		stackWin = 1
	elseif (cl.stackTile == 2)	// tile
		tileWin = 1
	endif

	CheckBox checkTile, win=$(panelName),value=tileWin
	CheckBox checkStack, win=$(panelName),value=stackWin
	
	
	// page orientation radio buttons
	Variable port, land
	if (cl.orientation == 0)		// portrait
		port = 1
	elseif (cl.orientation == 1)	// landscape
		land = 1
	endif
	CheckBox checkOrientation_port, win=$(panelName),value=port
	CheckBox checkOrientation_land, win=$(panelName),value=land
End

static Function GetRowInSelectedWave(windowName, cl)
	String windowName		// name of window to look for
	STRUCT createLayoutGlobals &cl

	FindValue/S=0/TEXT=windowName/TXOP=4/Z cl.selectedWindows
	
	return V_value
End

Function ParamChange_RowCol(cl)
	STRUCT createLayoutGlobals &cl
	
	// redimension the position text and select waves to match numRows and numCols
	Redimension/N=(cl.layoutRows, cl.layoutCols) cl.positionListText, cl.positionListSelect
	
	ParamChange_SelectedWindows(cl)
End

Function ParamChange_SelectedWindows(cl)
	// fills in the list box text wave selectedWindows with names of the graphs that
	// are selected in the window selection list box (windowListText/windowListSelect)
	STRUCT createLayoutGlobals &cl
	
	cl.positionListText = ""
	Variable numCells = min(numpnts(cl.positionListText), DimSize(cl.selectedWindows, 0))
	Variable n
	For (n=0; n<numCells; n+=1)
		cl.positionListText[floor(n/cl.layoutCols)][mod(n, cl.layoutCols)] = cl.selectedWindows[n]
	EndFor
End

Function CreateLayout(panelName)
	String panelName
	STRUCT createLayoutGlobals cl
	GetCreateLayoutsGlobals(cl, panelName)

	String layoutName = cl.stringTextWave[%LayoutName]
	String layoutTitle = cl.stringTextWave[%LayoutTitle]
	// see if LayoutName is in use
	String alertText
	DoWindow $(layoutName)
	if (V_Flag)		// window by that name already exists
		sprintf alertText, "A window with the name %s already exists.  Do you wish to overwrite the contents of that window?", layoutName
		DoAlert 2, alertText
		if (V_Flag == 1)		// yes clicked
			DoWindow/K $(layoutName)
		elseif (V_Flag == 2)	// no clicked
			// do nothing
		else					// cancel clicked
			return 0
		endif
	endif

	// use preferences, if user has checked that box
	if (cl.usePrefs)
		Preferences/Q 1	// turn preferences on
		Variable prefsOld = V_flag
	endif		
	
	String orientation = SelectString(cl.orientation, "Portrait", "Landscape")
	
	// create a layout
	if (cmpstr(layoutTitle, "") == 0)
		NewLayout/N=$(layoutName)/P=$(orientation)
	else
		NewLayout/N=$(layoutName)/P=$(orientation) as layoutTitle
	endif
	cl.stringTextWave[%LayoutName] = S_name	// store actual name in LayoutName variable
	layoutName = S_name
	
	// determine the origin and dimensions of the layout (all in points)
	Variable layoutX, layoutY, layoutWidth, layoutHeight
	String info = LayoutInfo(layoutName, "Layout")
	String pageValues = StringByKey("PAGE", info, ":", ";")
	layoutX = str2num(StringFromList(0, pageValues, ","))
	layoutY = str2num(StringFromList(1, pageValues, ","))
	layoutWidth = str2num(StringFromList(2, pageValues, ",")) - layoutX
	layoutHeight = str2num(StringFromList(3, pageValues, ",")) - layoutY
	
	// given grout, numCols, and numRows, determine the tileWidth and tileHeight
	Variable tileWidth, tileHeight
	tileWidth = (layoutWidth - ((cl.layoutCols - 1) * cl.groutThickness))/cl.layoutCols
	tileHeight = (layoutHeight - ((cl.layoutRows - 1) * cl.groutThickness))/cl.layoutRows
	
	// determine frameStyleNum
	Variable frameStyleNum = cl.frameStyleMode - 1		// frameStyleMode is 1 based; frameStyleNum is 0 based

	// add all of the objects to the layout.
	Variable n, numObjects = DimSize(cl.selectedWindows, 0)
	Variable left, top, right, bottom
	String winTypeStr
	For (n=0; n<numObjects; n+=1)
		// determine the type of window to be added
		Switch (WinType(cl.selectedWindows[n]))
			Case 1:		// graph
				winTypeStr = "graph"
				break
			Case 2:		// table
				winTypeStr = "table"
				break
			default:		// window type can't be used with AppendLayoutObject, so skip to next window
				continue
		EndSwitch
				
		// determine the left, top, right, and bottom positions for the object to be added
		Variable colNum, rowNum
		colNum = mod(n, cl.layoutCols)
		rowNum = floor(n/cl.layoutCols)

		Switch (cl.stackTile)
			Case 1:		// stack
				left = layoutX + (n * cl.groutThickness * 2)
				top = layoutY + (n * cl.groutThickness * 2)
				right = left + (layoutWidth / 3)
				bottom = top + (layoutHeight / 3)
				break
			default:		// tile by default
				left = layoutX + ((colNum * tileWidth) + (colNum * cl.groutThickness))
				top = layoutY + ((rowNum * tileHeight) + (rowNum  * cl.groutThickness))
				right = left + tileWidth
				bottom = top + tileHeight
		EndSwitch
	
		// add the object
		AppendLayoutObject /F=(frameStyleNum) /R=(left, top, right, bottom) /W=$(layoutName) $(winTypeStr)  $(cl.selectedWindows[n])
	EndFor
	
	if (cl.usePrefs)
		Preferences prefsOld
	endif
End

Function HandleWindowSelection(lba, cl)
	STRUCT WMListboxAction &lba
	STRUCT createLayoutGlobals &cl
	
	String graphName = lba.listWave[lba.row][%Name]
	if (WinType(graphName) != 0)		// if a window by this name exists
		Variable selectedWinWaveRow			// find what row in the selectedWindows wave this window name is, if it's there
		selectedWinWaveRow = GetRowInSelectedWave(graphName, cl)
		Variable numSelectedWaveRows = DimSize(cl.selectedWindows, 0)
		if (lba.selWave[lba.row][lba.col][0] & 0x10)		// did user just select the cell (and therefore window)
			if (selectedWinWaveRow < 0)		// window is not currently in selectedWindows, so add it at the end
				InsertPoints/M=0 numSelectedWaveRows, 1, cl.selectedWindows
				cl.selectedWindows[numSelectedWaveRows] = graphName
			endif
		else										// user just de-selected the cell (and therefore window)
			if (selectedWinWaveRow >= 0)		// window is currently in selectedWindows, so remove it
				DeletePoints/M=0 selectedWinWaveRow, 1, cl.selectedWindows
			endif
		endif
		ParamChange_SelectedWindows(cl)
	endif
End

//**
// Remove any windows that were previously selected/added to a layout
// but which are no longer available, because the user has closed the window,
// from the list of windows that the user may select to add to a layout.
//
// @ param cl
// 	A STRUCT reverence to the createLayoutGlobals structure that
// 	has already been initialized.
//*
static Function PurgeMissingWindows(cl)
	STRUCT createLayoutGlobals &cl
	
	Variable numSelected = DimSize(cl.selectedWindows, 0)
	Variable n
	
	String windowList = WinList("*", ";", "")
	String currentName
	Variable listItemNum
	Variable removedWindow		// Keeps track of whether or not any windows were removed from selectedWindows.
	// For each window found in selectedWindows, make sure it's still a valid window name.
	// Otherwise, remove it from selectedWindows
	For (n=0; n<numSelected; n+=1)
		currentName = cl.selectedWindows[n]
		listItemNum = WhichListItem(currentName, windowList, ";")
		if (listItemNum == -1)
			// Window is no longer available, so remove it
			// from the list of selected windows.
			DeletePoints n, 1, cl.selectedWindows
			removedWindow = 1
			
			// Since we just deleted a point in the wave, we need to reset
			// the value of n and also decrease numSelected by 1.
			n -= 1
			numSelected -= 1
		endif
	EndFor
	
	// Force a rebuild of all of the list box text/select waves if we've removed a window from selectedWindows
	if (removedWindow)
		ParamChange_SelectedWindows(cl)
	endif
End

//**
// Make sure a window is at as large or larger than a certain height and width,
// and if it is not then resize it.
//
// @param panelName
// 	A string containing the name of the window.
// @param minWidth
// 	The minimum allowed width of the target window, in points.
// @param minHeight
// 	The minimum allowed height of the target window, in points.
//*
Function MinWindowSize(windowName, minWidth, minHeight)
	String windowName
	Variable minWidth, minHeight

	GetWindow $(windowName) wsize
	Variable width= max(V_right-V_left, minWidth)
	Variable height= max(V_bottom-V_top, minHeight)
	MoveWindow/W=$(windowName) V_left, V_top, V_left+width, V_top+height
End

//**
// Repositions and resizes controls on the main CreateLayouts panel
// following resizing of that window by the user.
//
// @param panelName
// 	A string containing the name of the create layouts panel.
//*
Function ResizeMainPanelWindow(panelName)
	String panelName

	Variable outer_margin = 10
	Variable vertical_grout = 10
	Variable horizontal_grout = 20
	
	// Check to make sure that the window isn't too small.
	Variable min_width = kPANEL_MIN_LEFT_COL_WIDTH + (.5 * kPANEL_MIN_LEFT_COL_WIDTH) + 2 * (horizontal_grout + outer_margin)
	MinWindowSize(panelName,(min_width * 72 / ScreenResolution), (kPANEL_MIN_HEIGHT * 72 / ScreenResolution))
	
	// Get the new size of the window in points.
	GetWindow $(panelName) wsize
	Variable winHeight= V_bottom-V_top
	Variable winWidth = V_right-V_left
	
	// Convert point sizes into pixels.
	winHeight *= ScreenResolution/72	// points to pixels
	winWidth *= ScreenResolution/72	// points to pixels
	
	// Resize all of the components in the window as appropriate.
	//*****
	// Left half of panel
	//*****
			
	// First, reposition (but don't resize) the controls at the bottom left
	// of the panel (name, title, checkboxes, etc.) relative to the
	// bottom left corner of the window.

	Variable xpos, ypos = winHeight
	Variable ctrl_width, ctrl_height
	
	// Frame popup menu.
	xpos = outer_margin
	ypos -= outer_margin 
	RepositionControl(panelName, "popupFrameStyle", xpos, ypos, ctrl_width, ctrl_height, kNO_RESIZE)
	
	// Use layout prefs checkbox
	xpos += 135
	ypos += ctrl_height
	RepositionControl(panelName, "checkUsePrefs", xpos, ypos, ctrl_width, ctrl_height, kNO_RESIZE)
	
	ypos -= vertical_grout
	
	// Tile and stack checkboxes.
	xpos = outer_margin
	RepositionControl(panelName, "checkTile", xpos, ypos, ctrl_width, ctrl_height, kNO_RESIZE)
	xpos += 50
	ypos += ctrl_height
	RepositionControl(panelName, "checkStack", xpos, ypos, ctrl_width, ctrl_height, kNO_RESIZE)

	// Grout setvar.
	xpos += 55
	ypos += ctrl_height
	RepositionControl(panelName, "setvarGroutThickness", xpos, ypos, ctrl_width, ctrl_height, kNO_RESIZE)	

	// Name and title boxes.
	ypos -= vertical_grout
	xpos = outer_margin
	RepositionControl(panelName, "setvarTitle", xpos, ypos, ctrl_width, ctrl_height, kNO_RESIZE)
	ypos -= vertical_grout
	RepositionControl(panelName, "setvar0", xpos, ypos, ctrl_width, ctrl_height, kNO_RESIZE)
	
	// Store the y position of the top of the controls we've already repositioned.
	Variable y_hold = ypos

	// Page orientation group box and radios.
	xpos = outer_margin + 230
	ypos += 65
	RepositionControl(panelName, "groupOrientation", xpos, ypos, ctrl_width, ctrl_height, kNO_RESIZE)
	xpos += 15
	ypos += 34
	RepositionControl(panelName, "checkOrientation_port", xpos, ypos, ctrl_width, ctrl_height, kNO_RESIZE)
	ypos += 34
	RepositionControl(panelName, "checkOrientation_land", xpos, ypos, ctrl_width, ctrl_height, kNO_RESIZE)
	
	// Resize the window selection listbox to take up almost of the remaining
	// vertical space and about half of the horizontal space.
	ypos = y_hold - (2 * vertical_grout)
	ctrl_width = kPANEL_MIN_LEFT_COL_WIDTH
	ctrl_height = ypos - 35
	xpos = outer_margin
	RepositionControl(panelName, "listWindows", xpos, ypos, ctrl_width, ctrl_height, kBOTH_RESIZE)
	
	// Recenter the "Window Selection" text above the list box.
	ypos = outer_margin + 16
	Variable list_box_title_bottom = ypos
	xpos = outer_margin + (ctrl_width / 2) - 60
	RepositionControl(panelName, "titleWindowSelection", xpos, ypos, ctrl_width, ctrl_height, kNO_RESIZE)

	//*****
	// Right half of panel
	//*****
	Variable positioning_list_box_x_origin = outer_margin + kPANEL_MIN_LEFT_COL_WIDTH + (2 * horizontal_grout)

	// Reposition the Ok and Cancel buttons below the positioning list box.
	xpos = positioning_list_box_x_origin
	ypos = winHeight - outer_margin
	RepositionControl(panelName, "buttonCreateLayout", xpos, ypos, ctrl_width, ctrl_height, kNO_RESIZE)
	xpos += ctrl_width + horizontal_grout
	ypos += ctrl_height
	RepositionControl(panelName, "buttonCancel", xpos, ypos, ctrl_width, ctrl_height, kNO_RESIZE)
	
	// Reposition the rows and columns setvariables below the positioning list box.
	ypos -= 2 * vertical_grout
	xpos = positioning_list_box_x_origin
	RepositionControl(panelName, "setvarLayoutRows", xpos, ypos, ctrl_width, ctrl_height, kNO_RESIZE)
	xpos += ctrl_width + horizontal_grout
	ypos += ctrl_height
	RepositionControl(panelName, "setvarLayoutCols", xpos, ypos, ctrl_width, ctrl_height, kNO_RESIZE)
	
	// Resize and reposition the positioning list box.
	ypos = ypos - vertical_grout
	xpos = positioning_list_box_x_origin
	ctrl_width = winWidth - positioning_list_box_x_origin - outer_margin
	ctrl_height = ypos - 35
	RepositionControl(panelName, "listPosition", xpos, ypos, ctrl_width, ctrl_height, kBOTH_RESIZE)
	
	// Recenter the "Positioning" text above the list box.
	Variable positioning_list_box_x_center = positioning_list_box_x_origin + (ctrl_width / 2)
	ypos = list_box_title_bottom	
	xpos = positioning_list_box_x_center - 35
	RepositionControl(panelName, "titlePositioning", xpos, ypos, ctrl_width, ctrl_height, kNO_RESIZE)
	
	// Recalculate the row height of the list boxes containing graph and table icons.
	RecalculateListBoxRowHeight(panelName)
End

//**
// Repositions and possibly resizes a control on a window.
//
// @param windowName
// 	A string containing the name of a window.
// @param controlName
// 	A string containing the name of a control on the window.
// @param xpos
// 	The x position, in pixels, of the origin of the control.
// @param ypos
// 	The y position, in pixels, of the *bottom* of the control.
// 	The y position of the top of the control will be calculated
// 	by this function, and ypos will be replaced by this value
// 	after this function executes.  Passed by reference.
// @param ctrl_width
// 	The width of the control, in pixels.  When used as an
// 	input parameter, in conjunction with the resize parameter,
// 	this is the desired width, in pixels.  The actual width of the
// 	possibly resized control will be placed in this variable since
// 	it is passed by reference.
// @param ctrl_height
// 	The height of the control, in pixels.  When used as an
// 	input parameter, in conjunction with the resize parameter,
// 	this is the desired width, in pixels.  The actual height of the
// 	possibly resized control will be placed in this variable since
// 	it is passed by reference.
// @resize
// 	Whether or not this function should just reposition the control
// 	(resize == kNO_RESIZE) or reposition and resize the control.
// 	Pass kWIDTH_RESIZE to resize only the width, kHEIGHT_RESIZE
// 	to resize only the height, or kBOTH_RESIZE to resize both
// 	the height and width.
// *
Function RepositionControl(windowName, controlName, xpos, ypos, ctrl_width, ctrl_height, resize)
	String windowName
	String controlName
	Variable xpos
	Variable &ypos
	Variable &ctrl_width
	Variable &ctrl_height
	Variable resize
	
	// Don't do anything if windowName is not the name of a window.
	if (WinType(windowName) > 0)
		// Get current size of control
		ControlInfo/W=$(windowName) $(controlName)
		
		if (V_flag != 0)
			Variable new_height, new_width
			new_width = (resize &  kWIDTH_RESIZE) ? ctrl_width : V_width
			new_height = (resize & kHEIGHT_RESIZE) ? ctrl_height : V_height

			// Offset ypos based on the new height of the control
			ypos -= new_height
					
			// Reposition and possibly resize control
			ModifyControl/Z $(controlName) win=$(windowName), pos={xpos,ypos}, size={new_width, new_height}
			
			// Set value of output variables.
			ctrl_width = V_width
			ctrl_height = V_height
		endif
	endif	
End

//**
// Recalculate and resize row height of list box displaying
// icons with graphs and tables that shows the arrangement
// of those windows.
//
// @param windowName
// 	A string containing the name of the window
// 	that has the list box.
//*
Function RecalculateListBoxRowHeight(windowName)
	String windowName
	if (WinType(windowName) == 0)
		return 0
	endif

	Variable numRows, numCols
	Variable lb_height, lb_width
	Variable actual_row_height
		
	// Calculate the appropriate row height of the listPosition list box.
	ControlInfo/W=$(windowName) setvarLayoutRows
	if (abs(V_Flag) != 5)
		return 0
	endif
	numRows = V_value
	
	ControlInfo/W=$(windowName) setvarLayoutCols
	if (abs(V_Flag) != 5)
		return 0
	endif
	numCols = V_value
	
	// Get dimensions of the list box.
	ControlInfo/W=$(windowName) listPosition
	if (abs(V_Flag) != 11)
		return 0
	endif
	lb_height = V_height
	lb_width = V_width
	
	// Calculate the "ideal" row height.
	Variable ideal_row_height = V_height / numRows
	
	// If the ideal height falls between the minimum and maximum, then use it.
	// Otherwise, use the minimum or maximum height as appropriate.
	if ((ideal_row_height < kMAX_LB_ICON_HEIGHT) && (ideal_row_height > kMIN_LB_ICON_HEIGHT))
		actual_row_height = ideal_row_height
	elseif (ideal_row_height > kMAX_LB_ICON_HEIGHT)
		actual_row_height = kMAX_LB_ICON_HEIGHT
	else
		actual_row_height = kMIN_LB_ICON_HEIGHT
	endif

	// Resize the rows on the list box.
	ListBox listPosition win=$(windowName), special={1, actual_row_height, 0}
End

// ********************
//  HOOK FUNCTIONS
// ********************
Function Hook_MainPanel(str)
	STRUCT WMWinHookStruct &str
	
	STRUCT ACL_CreateLayoutsStr prefs
	STRUCT createLayoutGlobals cl
	GetCreateLayoutsGlobals(cl, str.winName)
	
	Variable statusCode = 0
	Switch (str.eventCode)
		Case 0:		// activate
			BuildWindowListWaves("*", kWINDOWTYPES, cl)		// rebuild the list of graphs
			PurgeMissingWindows(cl)
			ControlUpdate/A/W=$(str.winName)	
			break
		Case 2:		// window is being killed
			Kill_CreateLayouts(cl, str.winName)
			break
		Case 6:		// resize
			ResizeMainPanelWindow(str.winName)
			SaveDesktopControlPrefs(prefs)	// Save the new coordinates to the prefs file.
			statusCode=1
			break
		Case 12:	// window was moved: store new coordinates
			SaveDesktopControlPrefs(prefs)
			break
		default:			
	EndSwitch
	return statusCode
End


// ********************
//  PACKAGE PREFS LOADING/SAVING
// ********************
Function LoadDesktopControlPrefs(s)
	STRUCT ACL_CreateLayoutsStr &s
	
	// This loads preferences from disk if they exist on disk.
	LoadPackagePreferences kPackageName, kPreferencesFileName, kPreferencesRecordID, s

	// If error, prefs not found or not compatible, initialize them.
	if (V_flag!=0 || V_bytesRead==0 || s.version!=K_CURRENTPREFSVERSION)
		SaveDesktopControlPrefs(s)	// Create default prefs file.
	endif
	
	// code that loads prefs
End

Function SaveDesktopControlPrefs(s)
	STRUCT ACL_CreateLayoutsStr &s
	
	// fill in structure with current values
	s.version = K_CURRENTPREFSVERSION
	GetWindow $(kMAINPANELNAME) wsize			// get current window coordinates
	s.panelCoords[0] = V_left		// Left
	s.panelCoords[1] = V_top		// Top
	s.panelCoords[2] = V_right		// Right
	s.panelCoords[3] = V_bottom	// Bottom
	
	Variable i
	for(i=0; i<100; i+=1)
		s.reserved[i] = 0
	endfor

	SavePackagePreferences kPackageName, kPreferencesFileName, kPreferencesRecordID, s
	return 0	
End
